/* test.cpp
 *
 * Scott Livingston <SCL@utk.edu>
 * Originally created: 21 April 2009
 *
 */


#include <cv.h>
#include <highgui.h>

#include <libplayerc++/playerc++.h>
using namespace PlayerCc;

#include <cstdlib>
#include <iostream>
using namespace std;

#include "KernelObjTracker.hpp"
#include "ColorHistTracker.hpp"


#define NUM_KOT_OBJS 1


int main( int argc, char **argv )
{
	IplImage *img, *prevImg,
		*dispImage; // Use bounding boxes for visual feedback

	KernelObjTracker kot[NUM_KOT_OBJS];

	CvRect targetRegion[NUM_KOT_OBJS];

	int cam_width, cam_height, cam_depth;
	uint8_t *imgBuf;

	int i, j; // generic counters

	if (argc != 7) {
		cerr << "Usage: " << argv[0] << " init_image x0 y0 width height bandwidth" << endl;
		return 1;
	}

	PlayerClient robot( "localhost" );
	CameraProxy camp( &robot, 0 );

	// Determine camera image specs
	robot.Read();
	cam_width = camp.GetWidth();
	cam_height = camp.GetHeight();
	cam_depth = camp.GetDepth();

	// Attempt to load target initialization image
	if ((img = cvLoadImage( argv[1] )) == NULL) {
		cerr << "Error: Could not open image file: " << argv[1] << endl;
		return -1;
	}

	// Verify sign of given region arguments
	if (atoi(argv[2]) < 0 || atoi(argv[3]) < 0
		|| atoi(argv[4]) < 0 || atoi(argv[5]) < 0 ) {
		cerr << "Error: Given region specs (x0, y0, width, height) must be non-negative." << endl;
		return -1;
	}

	cvRectangle( img,
				 cvPoint(atoi(argv[2]), atoi(argv[3])),
				 cvPoint(atoi(argv[2])+atoi(argv[4]), atoi(argv[3])+atoi(argv[5])),
				 CV_RGB( 255, 0, 0 ), 5 );
	cvCircle( img, cvPoint(atoi(argv[2])+atoi(argv[4])/2,
						   atoi(argv[3])+atoi(argv[5])/2),
			  1, CV_RGB( 255, 0, 0 ), 5 );
	/*cvNamedWindow("track results");
	cvShowImage("track results", img );
	cvWaitKey(-1);
	return 0;*/
	
	// Attempt to instantiate and initialize a ColorHistTracker object
	ColorHistTracker cht( img, cvRect( atoi(argv[2]), atoi(argv[3]),
									   atoi(argv[4]), atoi(argv[5]) ),
						  16,     //number of histogram bins
						  .4,     //threshold
						  1024 ); //normalization factor

	// Attempt to instantiate and initialize a KernelObjTracker object
	/*KernelObjTracker kot( img, cvRect( atoi(argv[2]), atoi(argv[3]),
									   atoi(argv[4]), atoi(argv[5]) ),
						  16,            //number of histogram bins
						  atof(argv[6]), //bandwidth
						  false );       //adaptive bandwidth scaling?*/

	/* Scale initial target candidate location if necessary to handle
	   training images of different size than of the on-line camera
	   images. */
	/*kot.SetInitCenter( (atoi(argv[2])+atoi(argv[4])/2)*(float)(cam_width)/(img->width ? img->width : cam_width),
	  (atoi(argv[3])+atoi(argv[5])/2)*(float)(cam_height)/(img->height ? img->width : cam_height) );*/
	
	// Attempt to initialize a collection of KernelObjTracker objects
	for (i = 0; i < NUM_KOT_OBJS; i++) {
		kot[i].InitTargetModel( img, cvRect( atoi(argv[2]), atoi(argv[3]),
											 atoi(argv[4]), atoi(argv[5]) ),
								16,            //number of histogram bins
								atof(argv[6]), //bandwidth
								false );       //adaptive bandwidth scaling?*/
		if (i < NUM_KOT_OBJS/2)
			kot[i].SetInitCenter( cam_width/NUM_KOT_OBJS*i, 0 );
		else
			kot[i].SetInitCenter( cam_width/NUM_KOT_OBJS*i, cam_height/2 );
	}

	cvReleaseImage( &img ); // Init target image is no longer needed.
	img = NULL;

	// Image buffer and IplImage structure
	imgBuf = new uint8_t[cam_width*cam_height*cam_depth];
	img = cvCreateImage( cvSize(cam_width, cam_height), IPL_DEPTH_8U, 3 );

	// Open a window for visual feedback and enter main (tracking) loop
	cvNamedWindow( "track results", 1 );
	//cvNamedWindow( "prep", 1 );
	prevImg = NULL; // Indicate no previous image yet
	dispImage = NULL;

	cout << "Camera width, height, depth = " << cam_width << ", "
		 << cam_height << ", "
		 << cam_depth << endl;
	try {
		while (cvWaitKey(10) != 27) {

			robot.Read();
			camp.Decompress();
			camp.GetImage( imgBuf );

			// Copy image from camera proxy
			for (i = 0; i < cam_width; i++) {
				for (j = 0; j < cam_height; j++) {
					img->imageData[cam_width * j*3 + i*3 + 0] = (char)imgBuf[cam_width * j*3 + i*3 + 2];
					img->imageData[cam_width * j*3 + i*3 + 1] = (char)imgBuf[cam_width * j*3 + i*3 + 1];
					img->imageData[cam_width * j*3 + i*3 + 2] = (char)imgBuf[cam_width * j*3 + i*3 + 0];
				}
			}

			// Handle special case of first loop cycle
			if (prevImg == NULL) {
				prevImg = cvCloneImage( img );
				continue;
			}

			// Color histogram-based tracking
			//cht.GetMotionRegions( prevImg, img );
			//cht.MaxNCandidateRegions( 7 ); // Area-based pruning
			//cht.CompareRegions( img );
			//cht.CompareDerivRegions( img );
			//cht.OrderByArea(); //Area-based sorting of "target regions"
			//cht.PruneByArea( 1000 );
			//cht.OrderByCorrelation(); //Correlation magnitude-based sorting of "target regions"

			/* If high correlation region found with color
			   histogram-based tracker, then use it to initialize the
			   kernel-based object tracker. */
			/*if (cht.GetCount() > 0 && cht[0].r > .8
				&& ((cht[0].rect.x + cht[0].rect.width/2 - kot[0].GetCenterX())*(cht[0].rect.x + cht[0].rect.width/2 - kot[0].GetCenterX())
					+ (cht[0].rect.y + cht[0].rect.height/2 - kot[0].GetCenterY())*(cht[0].rect.y + cht[0].rect.height/2 - kot[0].GetCenterY()) > 1000)) {
				kot[0].SetInitCenter( cht[0].rect.x + cht[0].rect.width/2,
								   cht[0].rect.y + cht[0].rect.height/2 );
				cout << "UPDATED INIT CENTER r = " << cht[0].r << endl;
			}*/

			// Kernel-based object tracking 
			for (i = 0; i < NUM_KOT_OBJS; i++) {
				cout << "# " << i << " - distance: " << kot[i].Track( img ) << endl;
				kot[i].Track( img );
				targetRegion[i] = kot[i].GetTargetRegion();
			}
			cout << "---" << endl;

			// Draw rectangles on image to show found "target candidate"
			if (dispImage)
				cvReleaseImage( &dispImage );
			dispImage = cvCloneImage( img );
			for (i = NUM_KOT_OBJS-1; i >= 0; i--) {
				cvRectangle( dispImage,
							 cvPoint(targetRegion[i].x, targetRegion[i].y),
							 cvPoint(targetRegion[i].x+targetRegion[i].width, targetRegion[i].y+targetRegion[i].height),
							 CV_RGB( 255*(float)(NUM_KOT_OBJS-i)/NUM_KOT_OBJS, 0, 0 ), 5 );
				cvCircle( dispImage, cvPoint(targetRegion[i].x+targetRegion[i].width/2,
											 targetRegion[i].y+targetRegion[i].height/2),
						  1, CV_RGB( 255*(float)(NUM_KOT_OBJS-i)/NUM_KOT_OBJS, 0, 0 ), 5 ); // Mark center of target candidate region
			}
			cvShowImage( "track results", dispImage );

			// Prepare for next loop cycle
			cvCopy( img, prevImg );
		}
	} catch (PlayerError e) {
		cerr << e << endl;
		return -1;
	}

	cvDestroyWindow( "track results" );
	//cvDestroyWindow( "prep" );

	// Release image buffer and structures
	if (img != NULL)
		cvReleaseImage( &img );
	if (prevImg != NULL)
		cvReleaseImage( &prevImg );
	delete [] imgBuf;

	return 0;
}
